from EasyIPCLarge.Msg import Msg
from thrift.transport import TSocket
from thrift.server import TNonblockingServer
from threading import Thread
from typing import Callable, Dict
import time
import json


class MsgHandler:
    def __init__(self,
                 OnReq: Callable[[Dict], Dict] = None):
        self.__OnReq = OnReq

    def Send(self, req):
        if self.__OnReq:
            return json.dumps(self.__OnReq(json.loads(req)))
        else:
            return json.dumps({})


class MsgServer:
    def __init__(self,
                 port:  int,
                 OnReq: Callable[[Dict], Dict] = None):
        handler = MsgHandler(OnReq)
        processor = Msg.Processor(handler)
        transport = TSocket.TServerSocket("127.0.0.1", port)
        self.__server = TNonblockingServer.TNonblockingServer(processor,
                                                              transport)
        self.__t = Thread(target=self.__Loop)
        self.__t.start()

    def __Loop(self):
        self.__server.serve()

    def Stop(self):
        if self.__t:
            self.__server.stop()
            self.__t.join()


if __name__ == "__main__":
    server = MsgServer(8000)
    time.sleep(24*3600)
    server.Stop()



from EasyIPCLarge.MsgClient import MsgClient
from EasyIPCLarge.NameConst import NameConst as nc
from multiprocessing import shared_memory
from threading import Thread
from typing import Dict
import uuid
import threading
import time


class MemoryClient:
    def __init__(self,
                 port: int):
        self.__msgClient = MsgClient(port)
        self.__sessionID = str(uuid.uuid4())
        self.__running = False
        self.__memory = None
        self.__memorySize = 0
        self.__memoryIndexR = 0
        self.__memoryIndexW = 0
        self.__memoryBlocks = []
        self.__memoryThread = None
        self.__gCond = threading.Condition()

    def __del__(self):
        self.__CleanShareMemory()

    def __PrepareShareMemory(self, size: int):
        self.__CleanShareMemory()
        self.__memory = shared_memory.SharedMemory(create=True, size=size)
        self.__memorySize = size
        self.__memoryIndexR = 0
        self.__memoryIndexW = 0

    def __CleanShareMemory(self):
        if self.__memory is not None:
            self.__memory.close()
            self.__memory.unlink()
            self.__memory = None

    def SharedMemoryOpen(self, memorySize: int):
        self.__PrepareShareMemory(memorySize)

        req = {}
        req[nc.kSessionID] = self.__sessionID
        req[nc.kAction] = nc.vOpen
        req[nc.kSharedMemoryName] = self.__memory.name
        self.__msgClient.Req(req)

        self.__memoryThread = Thread(target=self.__SharedMemoryNotify)
        self.__memoryThread.start()

    def SharedMemoryClose(self):
        self.__running = False
        self.__gCond.acquire()
        self.__gCond.notifyAll()
        self.__gCond.release()
        if self.__memoryThread:
            self.__memoryThread.join()
            self.__memoryThread = None

        req = {}
        req[nc.kSessionID] = self.__sessionID
        req[nc.kAction] = nc.vClose
        self.__msgClient.Req(req)

        self.__CleanShareMemory()

    def SharedMemorySend(self,
                         seq: int,
                         data: bytes,
                         timeout: float = 1.0) -> bool:
        ret = False
        sz = len(data)
        t0 = time.time_ns() // 1000

        self.__gCond.acquire()

        ind = self.__SharedMemoryRequireSpace(sz)
        if ind < 0:
            self.__gCond.wait(timeout)
            ind = self.__SharedMemoryRequireSpace(sz)

        if ind >= 0:
            self.__memory.buf[ind:ind+sz] = data[:]
            self.__memoryIndexW += sz
            self.__memoryBlocks.append([ind, sz, seq, t0])
            self.__gCond.notifyAll()
            ret = True

        self.__gCond.release()

        return ret

    def SharedMemoryEmpty(self) -> bool:
        return len(self.__memoryBlocks) == 0

    def ExternalReq(self, req: Dict) -> Dict:
        reqTpl = {}
        reqTpl[nc.kSessionID] = self.__sessionID
        reqTpl[nc.kAction] = nc.vExternal
        reqTpl[nc.kTime] = str(time.time_ns() // 1000)
        reqTpl[nc.kBody] = req
        respTpl = self.__msgClient.Req(reqTpl)
        return respTpl[nc.kBody]

    def ResetServerSeq(self):
        req = {}
        req[nc.kSessionID] = self.__sessionID
        req[nc.kAction] = nc.vResetSeq
        self.__msgClient.Req(req)

    def __SharedMemoryRequireSpace(self, sz: int) -> int:
        indR = self.__memoryIndexR
        indW = self.__memoryIndexW
        if indW >= indR:
            if indW + sz < self.__memorySize:
                return indW
            else:
                if sz < indR:  # rewind
                    self.__memoryIndexW = 0
                    return 0
                else:
                    return -1
        else:
            if indR - indW > sz:
                return indW
            else:
                return -1

    def __SharedMemoryNotify(self):
        self.__running = True
        wTimeLong = 0.1
        wTimeShort = 0.002

        while self.__running:
            self.__gCond.acquire()
            while self.__running and len(self.__memoryBlocks) == 0:
                self.__gCond.wait(wTimeLong)
            self.__gCond.release()

            if not self.__running:
                break

            info = self.__memoryBlocks[0]
            ind = info[0]
            sz = info[1]
            seq = info[2]
            t0 = info[3]
            req = {}
            req[nc.kSessionID] = self.__sessionID
            req[nc.kAction] = nc.vSend
            req[nc.kRange] = "{}-{}".format(ind, ind + sz)
            req[nc.kSeq] = str(seq)
            req[nc.kTime] = str(t0)

            while self.__running:
                resp = self.__msgClient.Req(req)
                if resp[nc.kErr] == nc.vErrOK:
                    self.__memoryIndexR = (ind + sz) % self.__memorySize
                    self.__memoryBlocks.pop(0)
                    self.__gCond.acquire()
                    self.__gCond.notifyAll()
                    self.__gCond.release()
                    break
                time.sleep(wTimeShort)



class NameConst:
    kSessionID = "SessionID"

    kAction   = "Action"    # noqa
    vOpen     = "Open"      # noqa
    vClose    = "Close"     # noqa
    vSend     = "Send"      # noqa
    vExternal = "External"
    vResetSeq = "ResetSeq"

    kShareType    = "ShareType"  # noqa
    vSharedMemory = 1
    vSocket       = 2            # noqa

    kSharedMemoryName = "SharedMemoryName"

    kRange = "Range"
    kSeq   = "Seq"    # noqa
    kTime  = "Time"   # noqa
    kBody  = "Body"   # noqa

    kErr         = "Err"        # noqa
    vErrOK       = "OK"         # noqa
    vErrNoMemory = "No memory"



from EasyIPCLarge.Msg import Msg
from thrift.transport import TSocket
from thrift.transport import TTransport
from thrift.protocol import TBinaryProtocol
from typing import Dict
import json


class MsgClient:
    def __init__(self,
                 port:  int):
        sock = TSocket.TSocket("127.0.0.1", port)
        self.__transport = TTransport.TFramedTransport(sock)
        protocol = TBinaryProtocol.TBinaryProtocol(self.__transport)
        self.__client = Msg.Client(protocol)
        self.__transport.open()

    def Req(self,
            req: Dict
            ) -> Dict:
        return json.loads(self.__client.Send(json.dumps(req)))

    def __del__(self):
        self.__transport.close()


if __name__ == "__main__":
    m = MsgClient(8000)



from EasyIPCLarge.MsgServer import MsgServer
from EasyIPCLarge.NameConst import NameConst as nc
from multiprocessing import shared_memory
from typing import Callable, Dict, Tuple
import threading
import time


class DataSession:
    def __init__(self, sessionID: str, sharedMemoryName: str):
        self.__sessionID = sessionID
        self.__sharedMemoryName = sharedMemoryName
        self.__memory = shared_memory.SharedMemory(name=sharedMemoryName)

    def __del__(self):
        if self.__memory:
            self.__memory.close()

    @property
    def memory(self):
        return self.__memory


class MemoryServer:
    def __init__(self,
                 port: int,
                 memoryLimit: int = 32*1024*1024,
                 onExternalReq: Callable[[int, Dict], Dict] = None):
        self.__msgServer = MsgServer(port, OnReq=self.__OnReq)
        self.__onExternalReq = onExternalReq
        self.__sessions = {}
        self.__seqNext = 0
        self.__blocks = []
        self.__memoryFreeBytes = memoryLimit
        self.__gCond = threading.Condition()

    def Stop(self):
        self.__gCond.acquire()
        self.__gCond.notifyAll()
        self.__gCond.release()
        if self.__msgServer:
            self.__msgServer.Stop()

    def NoSessions(self) -> bool:
        return len(self.__sessions) == 0

    def NoData(self) -> bool:
        return len(self.__blocks) == 0

    def GetData(self, timeout: float = 1.0) -> Tuple[int, bytes]:
        seq, b = -1, None
        tCreate = None

        self.__gCond.acquire()
        if len(self.__blocks) == 0 or self.__blocks[0][0] != self.__seqNext:
            self.__gCond.wait(timeout)
        if len(self.__blocks) != 0 and self.__blocks[0][0] == self.__seqNext:
            block = self.__blocks.pop(0)
            seq = block[0]
            tCreate = block[1]
            b = block[2]
            self.__seqNext += 1
            self.__memoryFreeBytes += len(b)
        self.__gCond.release()

        delayUS = 0 if tCreate is None else time.time_ns() // 1000 - tCreate
        return seq, delayUS, b

    def __OnReq(self, req: Dict) -> Dict:
        resp = {}
        resp[nc.kErr] = nc.vErrOK

        sessionID = req.get(nc.kSessionID)
        action = req.get(nc.kAction)

        if action == nc.vOpen:
            sharedMemoryName = req.get(nc.kSharedMemoryName)
            self.__sessions[sessionID] = DataSession(sessionID,
                                                     sharedMemoryName)
        elif action == nc.vClose:
            session = self.__sessions.pop(sessionID)
            del session
        elif action == nc.vSend:
            session = self.__sessions.get(sessionID)
            if session and session.memory:
                seq = int(req[nc.kSeq])
                timeCreate = int(req[nc.kTime])
                ranges = req[nc.kRange].split("-")
                indL = int(ranges[0])
                indR = int(ranges[1])
                if not self.__BlockInsert(seq, timeCreate, session.memory.buf,
                                          indL, indR):
                    resp[nc.kErr] = nc.vErrNoMemory
        elif action == nc.vExternal:
            t0 = int(req[nc.kTime])
            delayUS = time.time_ns() // 1000 - t0
            if self.__onExternalReq:
                resp[nc.kBody] = self.__onExternalReq(delayUS, req[nc.kBody])
        elif action == nc.vResetSeq:
            self.__seqNext = 0

        return resp

    def __BlockInsert(self, seq: int, timeCreate: int,
                      buf: bytes, indL: int, indR: int) -> bool:
        if (self.__memoryFreeBytes < indR - indL) and (seq != self.__seqNext):
            return False

        b = bytes(buf[indL:indR])
        self.__gCond.acquire()
        self.__blocks.append([seq, timeCreate, b])
        self.__memoryFreeBytes -= len(b)
        self.__blocks.sort(key=lambda x: x[0])
        self.__gCond.notifyAll()
        self.__gCond.release()

        return True



import setuptools


with open("README.md", "r") as f:
    long_description = f.read()


setuptools.setup(
    name="EasyIPCLarge",
    version="1.0.0",
    author="YunWen",
    author_email="yunwen@xiaobing.ai",
    description="A library for high-speed transfer of large data between processes",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    license="MIT",
    python_requires=" >= 3.8",
    install_requires=[
        'thrift >= 0.16.0',
    ]
)
